// =====================================================================
// 
// Copyright (C) 2000 - 2020 
//   Charles Brain <chbrain@dircon.co.uk>
//   Ilkka Toivanen <pile@aimo.kareltek.fi>
//   David Freese <w1hkj@bellsouth.net>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// File: ale.cxx
//
// =====================================================================

#include <iostream>

#include "configuration.h"
#include "confdialog.h"
#include "status.h"
#include "debug.h"
#include "FTextRXTX.h"
#include "modem.h"
#include "qrunner.h"
#include "main.h"
#include "rigio.h"

#include "fl_digi.h"
#include "ale.h"
#include "ale_codec.h"
#include "ale_config.h"
#include "ale_dialog.h"

namespace ALE {

static pthread_t       ale_pthread;
static pthread_cond_t  ale_cond;

static pthread_mutex_t ale_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t txt_mutex = PTHREAD_MUTEX_INITIALIZER;

#define ALE_LOOP_DELAY  10 // in milliseconds

static int scan_wait = 1000 / ALE_LOOP_DELAY;  // 1 second start up

std::list<str_text> text_list;
std::list<std::string> ale_reports_list;

int sp_compare(std::string s1, std::string s2)
{
	float f1, f2;
	f1 = atof(s1.c_str());
	f2 = atof(s2.c_str());
	if (f1 < f2) return -1;
	if (f2 > f1) return 1;
	return 0;
}

void init_scn_frequencies ()
{
	std::string alestr = ale_config.scn_frequencies;
	size_t p1 = alestr.find(":");
	size_t p2 = 0;

	list<std::string> scps;
	std::string sp;

	while (p1 != std::string::npos) {
		if (p1 != std::string::npos) {
			sp = alestr.substr (0, p1).append(":").append (alestr[p1 + 1] == '1' ? "1" : "0");
			scps.push_back(sp);
		}
		p2 = alestr.find(",");
		if (p2 != std::string::npos)
			alestr.erase (0, p2 + 1);
		else {
			alestr.erase(0);
			break;
		}
		p1 = alestr.find (":");
	}
	p1 = alestr.find(":");
	if (p1 != std::string::npos) {
		sp = alestr.substr (0, p1).append(":").append (alestr[p1 + 1] == '1' ? "1" : "0");
		scps.push_back(sp);
	}

	scps.sort(sp_compare);
	scn_frequencies->clear();
	size_t p;
	while (!scps.empty()) {
		sp = scps.front();
		p = sp.find(":");
		scn_frequencies->add (sp.substr(0,p).c_str(), sp[p+1]=='1' ? 1 : 0);
		scps.pop_front();
	}

}

void init_snd_frequencies()
{
	std::string alestr = ale_config.snd_frequencies;
	size_t p1 = alestr.find(":");
	size_t p2 = 0;

	list<std::string> scps;
	std::string sp;

	while (p1 != std::string::npos) {
		if (p1 != std::string::npos) {
			sp = alestr.substr (0, p1).append(":").append (alestr[p1 + 1] == '1' ? "1" : "0");
			scps.push_back(sp);
		}
		p2 = alestr.find(",");
		if (p2 != std::string::npos)
			alestr.erase (0, p2 + 1);
		else {
			alestr.erase(0);
			break;
		}
		p1 = alestr.find (":");
	}
	p1 = alestr.find(":");
	if (p1 != std::string::npos) {
		sp = alestr.substr (0, p1).append(":").append (alestr[p1 + 1] == '1' ? "1" : "0");
		scps.push_back(sp);
	}

	scps.sort(sp_compare);
	snd_frequencies->clear();
	size_t p;
	while (!scps.empty()) {
		sp = scps.front();
		p = sp.find(":");
		snd_frequencies->add (sp.substr(0,p).c_str(), sp[p+1]=='1' ? 1 : 0);
		scps.pop_front();
	}

}

void init_frequencies ()
{
	ALE::init_scn_frequencies ();
	ALE::init_snd_frequencies ();
}

void update_scn_frequencies()
{
	if (Fl::event_state() & FL_SHIFT) {
		int n = scn_frequencies->value();
		scn_frequencies->remove(n);
	}
	std::string f_items;
	for (int n = 1; n <= scn_frequencies->nitems(); n++) {
		if (n > 1) f_items.append(",");
		f_items.append (scn_frequencies->text(n)).append(":");
		f_items.append (scn_frequencies->checked(n) ? "1" : "0");
	}
	ale_config.scn_frequencies = f_items;
}

void add_scan_frequency()
{
	float f;
	std::string szf = inp_scan_frequency->value();
	if (szf.empty()) return;

	int result = sscanf(szf.c_str(), "%f", &f);
	if (!result) return;

	if (scn_frequencies->nitems() == 0) {
		scn_frequencies->add (szf.c_str(), 1);
		return;
	}
	list<std::string> scps;
	std::string sp;
	for (int n = 1; n <= scn_frequencies->nitems(); n++) {
		sp = scn_frequencies->text(n);
		sp.append(":").append(scn_frequencies->checked(n) ? "1" : "0");
		scps.push_back(sp);
	}
	sp = szf.append(":1");
	scps.push_back(sp);
	scps.sort(sp_compare);
	scn_frequencies->clear();
	size_t p;
	while (!scps.empty()) {
		sp = scps.front();
		p = sp.find(":");
		scn_frequencies->add (sp.substr(0,p).c_str(), sp[p + 1]=='1' ? 1 : 0);
		scps.pop_front();
	}
}

void scan_frequencies()
{
	if (Fl::event_state() & FL_SHIFT) {
		int n = snd_frequencies->value();
		snd_frequencies->remove(n);
	}
	std::string f_items;
	for (int n = 1; n <= snd_frequencies->nitems(); n++) {
		if (n > 1) f_items.append(",");
		f_items.append (snd_frequencies->text(n)).append(":");
		f_items.append (snd_frequencies->checked(n) ? "1" : "0");
	}
	ale_config.snd_frequencies = f_items;
}

void update_snd_frequencies()
{
	if (Fl::event_state() & FL_SHIFT) {
		int n = snd_frequencies->value();
		snd_frequencies->remove(n);
	}
	std::string f_items;
	for (int n = 1; n <= snd_frequencies->nitems(); n++) {
		if (n > 1) f_items.append(",");
		f_items.append (snd_frequencies->text(n)).append(":");
		f_items.append (snd_frequencies->checked(n) ? "1" : "0");
	}
	ale_config.snd_frequencies = f_items;
}

void add_snd_frequency()
{
	float f;
	std::string szf = inp_snd_frequency->value();
	if (szf.empty()) return;

	int result = sscanf(szf.c_str(), "%f", &f);
	if (!result) return;

	if (snd_frequencies->nitems() == 0) {
		snd_frequencies->add (szf.c_str(), 1);
		return;
	}
	list<std::string> scps;
	std::string sp;
	for (int n = 1; n <= snd_frequencies->nitems(); n++) {
		sp = snd_frequencies->text(n);
		sp.append(":").append(snd_frequencies->checked(n) ? "1" : "0");
		scps.push_back(sp);
	}
	sp = szf.append(":1");
	scps.push_back(sp);
	scps.sort(sp_compare);
	snd_frequencies->clear();
	size_t p;
	while (!scps.empty()) {
		sp = scps.front();
		p = sp.find(":");
		snd_frequencies->add (sp.substr(0,p).c_str(), sp[p + 1]=='1' ? 1 : 0);
		scps.pop_front();
	}
}

static void write_rxtx()
{
	str_text text_pair;
	static std::string heard;
	guard_lock txtlock(&txt_mutex);
	while (!text_list.empty()) {
		text_pair = text_list.front();
		text_list.pop_front();
		ReceiveText->addstr(text_pair.text.c_str(), text_pair.style);
		if (text_pair.style == FTextBase::XMIT)
			txt_ale_sent->addstr(text_pair.text);
	}
	while (!ale_reports_list.empty()) {
		heard = ale_reports_list.front();
		ale_reports_list.pop_front();
		if (ale_config.rx_topline) {
			ale_reports->insert(1, heard.c_str());
			ale_reports->topline(1);
		} else {
			ale_reports->add(heard.c_str());
			ale_reports->topline(ale_reports->size());
		}
	}
}

void write_rxstr(std::string rx_str, int ber, int sinad)
{
	guard_lock txtlock(&txt_mutex);
	static char msgout[100];
	snprintf (msgout,
			  sizeof(msgout),
			  "%s,%6d,%6d,%6d\n",
			  rx_str.c_str(),
			  ber,
			  int(sinad),
			  48 - ber + int(sinad));

	str_text rxstr;

	rxstr.text = msgout;
	rxstr.style = FTextBase::RECV;
	text_list.push_back(rxstr);
	ale_reports_list.push_back(msgout);
	REQ(write_rxtx);

	if (progStatus.LOGenabled)
		logfile->log_to_file(cLogfile::LOG_RX, msgout);
}

void write_txstr(std::string tx_string)
{
	guard_lock txtlock(&txt_mutex);
	str_text txstr;
	txstr.text = tx_string;
	txstr.style = FTextBase::XMIT;
	text_list.push_back(txstr);
	REQ(write_rxtx);

	if (progStatus.LOGenabled)
		logfile->log_to_file(cLogfile::LOG_TX, tx_string);

}

void sort_rx()
{
	guard_lock txtlock(&txt_mutex);
	std::list<std::string> lines;
	int nlines = ale_reports->size();
	if (!nlines) return;
	for (int i = 1; i <= nlines; i++)
		lines.push_back(ale_reports->text(i));
	lines.sort();
	ale_reports->clear();
	for (int i = 0; i < nlines; i++) {
		if (ale_config.rx_topline) {
			ale_reports->add(lines.back().c_str());
			lines.pop_back();
		} else {
			ale_reports->add(lines.front().c_str());
			lines.pop_front();
		}
	}
	if (ale_config.rx_topline) {
		ale_reports->topline(1);
	} else {
		ale_reports->topline(ale_reports->size());
	}
}

pthread_mutex_t heard_mutex = PTHREAD_MUTEX_INITIALIZER;
std::vector<std::string> ale_heard_list;

void update_heard()
{
	guard_lock heard_lock(&heard_mutex);
	static std::string line;
	ale_heard->clear();
	for (size_t n = 0; n < ale_heard_list.size(); n++) {
		line = ale_heard_list[n];
		ale_heard->add(line.c_str());
	}
	ale_heard_list.clear();
	ale_heard->topline(ale_reports->size());
	ale_heard->redraw();
}

// line: who, freq, LQA, ber, sinad, dtime
static void split_fields(std::string line, heard_fields &fields)
{
	if (line.empty()) return;
	size_t p = line.find (", "); // points to freq
	fields.preamble = 0;//atol (line.c_str());
	fields.who = line.substr(0, p);
	line.erase (0, p + 2);
	p = line.find (", ");
	line.erase (0, p + 2); // points to LQA
	fields.LQA = atol (line.c_str());
	p = line.find (", "); // points to ber entry
	line.erase (0, p + 2);
	p = line.find (", "); // points to sinad entry
	line.erase (0, p + 2);
	p = line.find (", "); // points to dtime entry
	line.erase(0, p + 2);
	fields.dtime = line;
}

static long dt2long(std::string dtstr)
{
	if (dtstr.empty()) return 0;
	long val = 0;
	size_t p;
	val = atol (dtstr.c_str()) - 1900;
	val *= 365;
	p = dtstr.find (" ");
	dtstr.erase (0, p + 1);
	val += atol (dtstr.c_str());
	val *= 24;
	p = dtstr.find (" ");
	dtstr.erase (0, p + 1);
	val += atol (dtstr.substr(0,2).c_str());
	val *= 60;
	dtstr.erase (0,2);
	val += atol (dtstr.c_str());
	return val;
}

static void age_ale_heard_list()
{
	if (ale_heard_list.empty()) return;

	std::string  line1, line2;
	list<std::string> culled;
	heard_fields fields1, fields2;

	long now;
	time_t timestamp = time(NULL);
	struct tm *UTC_time = gmtime(&timestamp);
	now = UTC_time->tm_year * 365;
	now += UTC_time->tm_yday;
	now *= 24;
	now += UTC_time->tm_hour;
	now *= 60;
	now += UTC_time->tm_min;

	sort(ale_heard_list.begin(), ale_heard_list.end());

	for (size_t n = 0; n < ale_heard_list.size(); n++) {
		line2 = ale_heard_list[n];
		split_fields(line2, fields2);
		if (n == 0) {
			line1 = line2;
			fields1 = fields2;
			continue;
		}
		if ((fields2.who != fields1.who) ||
			((fields2.who == fields1.who) && (fields2.LQA < fields1.LQA))) {
			if (now - dt2long(fields1.dtime) < ale_config.aging)
				culled.push_back(line1);
		}
		line1 = line2;
		fields1 = fields2;

	}
	if (now - dt2long(fields1.dtime) < ale_config.aging)
		culled.push_back(line1);

	ale_heard_list.clear();
	while (!culled.empty()) {
		ale_heard_list.push_back(culled.front());
		culled.pop_front();
	}
}

static ALE::heard_fields last_heard;
static ALE::heard_fields this_heard;

void parse_TIS_TWS (int preamble, std::string who, int ber, int sinad)
{
	char szdt[20];
	time_t timestamp = time(NULL);
	struct tm *UTC_time = gmtime(&timestamp);
	snprintf(szdt, sizeof(szdt), "%4d %03d %02d%02d",
		1900 + UTC_time->tm_year, UTC_time->tm_yday, UTC_time->tm_hour, UTC_time->tm_min);

//	if ((this_heard.who == last_heard.who && 
//		 this_heard.LQA >= last_heard.LQA ) ||
//		this_heard.who != last_heard.who) {

		char szbuff[200];
		last_heard = this_heard;
		memset(szbuff, '\0', 200);
		snprintf (szbuff, sizeof(szbuff),
			"%s, %0.1f, %d, %d, %d, %s",
			who.c_str(),
			qsoFreqDisp->value() * 1e-3,
			48 - ber + sinad,
			ber,
			sinad,
			szdt);

		guard_lock heard_lock(&heard_mutex);
		int nlines = ale_heard->size();
		ale_heard_list.clear();
		if (nlines) {
			for (int i = 1; i <= nlines; i++)
				ale_heard_list.push_back(ale_heard->text(i));
		}
		ale_heard_list.push_back(szbuff);
		age_ale_heard_list();

		REQ(update_heard);
//	}
}

void parse_heard (int preamble, std::string who, int ber, int sinad)
{
	switch (preamble) {
		case ale_codec::TIS : 
		case ale_codec::TWS : parse_TIS_TWS(preamble, who, ber, sinad); break;
		case ale_codec::DAT : break;
		case ale_codec::THRU: break;
		case ale_codec::TO : break;
		case ale_codec::FROM : break;
		case ale_codec::CMD : break;
		case ale_codec::REP : break;
	}
	dwell_on_channel();
}

void sort_heard()
{
	{
	guard_lock heard_lock(&heard_mutex);
		int nlines = ale_heard->size();
		if (!nlines) return;
		ale_heard_list.clear();
		for (int i = 1; i <= nlines; i++)
			ale_heard_list.push_back(ale_heard->text(i));
		sort(ale_heard_list.begin(), ale_heard_list.end());
	}
	update_heard();
}

//----------------------------------------------------------------------
static bool ale_thread_running   = false;
static bool ale_terminate_flag   = false;
//----------------------------------------------------------------------

std::string ale_str;

bool enable_scanning = false;

static void do_qsy(void *data)
{
	size_t rfc = (size_t)data;
	if (progdefaults.chkUSERIGCATis)
		rigCAT_set_qsy (rfc);
#if USE_HAMLIB
	else if (progdefaults.chkUSEHAMLIBis)
		hamlib_set_qsy (rfc);
#endif
	else
		qso_selectFreq(rfc, 1500);
}

void qsy(long rfc)
{
	if (connected_to_flrig)
		xmlrpc_rig_set_qsy (rfc);
	else
		Fl::awake(do_qsy, (void *)rfc);
}

static void scan_next()
{
	static int next = 1;
	int last = scn_frequencies->nitems();

	if (last == 0) return;

	if (next > last) next = 1;
	while (next <= last) {
		if (scn_frequencies->checked(next)) { // scan next
			long freq = 1000 * atof(scn_frequencies->text(next));
			qsy (freq);
			if (++next > last) next = 1;
			return;
		} else 
			if (++next > last) next = 1;
		if (scn_frequencies->nchecked() == 0) return;
		if ( (last = scn_frequencies->nitems()) == 0) return;
	}
}

void dwell_on_channel()
{
	guard_lock loop_lock(&ale_mutex);
	scan_wait += (ale_config.dwell_time / ALE_LOOP_DELAY);
}

static void * ale_loop(void *args) // executes scanning function
{
	SET_THREAD_ID(ALE_TID);

	static int scndur[] = { 1000, 500, 333, 250, 200 };

	ale_thread_running   = true;
	ale_terminate_flag   = false;

	while(1) {

//		pthread_mutex_lock(&ale_mutex);
//		pthread_cond_wait(&ale_cond, &ale_mutex);
//		pthread_mutex_unlock(&ale_mutex);

		MilliSleep(ALE_LOOP_DELAY);

		if (ale_terminate_flag)
			break;

		{
			guard_lock looplock (&ale_mutex);
			if (enable_scanning && --scan_wait == 0) {
				scan_next();
				scan_wait = scndur[ale_config.scan_rate] / ALE_LOOP_DELAY;
			}
		}
//		send_ale(ale_str);
	}
	return (void *)0;
}

void pause_scanning()
{
	guard_lock looplock (&ale_mutex);
	enable_scanning = false;
}

void continue_scanning()
{
	guard_lock looplock (&ale_mutex);
	enable_scanning = true;
}

void stop_ale_thread(void)
{
	if(!ale_thread_running) return;

	ale_terminate_flag = true;
	pthread_cond_signal(&ale_cond);

	MilliSleep(10);

	pthread_join(ale_pthread, NULL);

	pthread_mutex_destroy(&ale_mutex);
	pthread_cond_destroy(&ale_cond);

	memset((void *) &ale_pthread, 0, sizeof(ale_pthread));
	memset((void *) &ale_mutex,   0, sizeof(ale_mutex));

	ale_thread_running   = false;
	ale_terminate_flag   = false;
}

void start_ale_thread(void)
{
	if (ale_thread_running) return;

	memset((void *) &ale_pthread, 0, sizeof(ale_pthread));
	memset((void *) &ale_mutex,   0, sizeof(ale_mutex));
	memset((void *) &ale_cond,    0, sizeof(ale_cond));

	if(pthread_cond_init(&ale_cond, NULL)) {
		LOG_ERROR("ALE thread create fail (pthread_cond_init)");
		return;
	}

	if(pthread_mutex_init(&ale_mutex, NULL)) {
		LOG_ERROR("ALE thread create fail (pthread_mutex_init)");
		return;
	}

	if (pthread_create(&ale_pthread, NULL, ale_loop, NULL) < 0) {
		pthread_mutex_destroy(&ale_mutex);
		LOG_ERROR("ALE thread create fail (pthread_create)");
	}

	LOG_INFO("started ALE thread");

	MilliSleep(10); // Give the CPU time to set 'ale_thread_running'
}

} // namespace ALE
