// =====================================================================
//
// Copyright (C) 2000 - 2020
//   Charles Brain <chbrain@dircon.co.uk>
//   Ilkka Toivanen <pile@aimo.kareltek.fi>
//   David Freese <w1hkj@bellsouth.net>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// File: ale_codec.cc
//
// =====================================================================

#include <iostream>

#include "ale.h"
#include "ale_codec.h"
#include "ale_dialog.h"
#include "ale_config.h"

#include "misc.h"
#include "configuration.h"
#include "confdialog.h"
#include "status.h"
#include "debug.h"
#include "FTextRXTX.h"
#include "modem.h"
#include "qrunner.h"
#include "main.h"

#include "threads.h"

#define min(a,b) ((a) < (b) ? a : b)
#define max(a,b) ((a) > (b) ? a : b)

string ale_codec::preamble_strs[] = {
	"DATA", // extension of data field of the previous ALE word
	"THRU", // multiple direct destinations for group calls
	"TO",   // direct destination for individual or net calls
	"TWS",  // id of present sender, signal & protocall termination
	"FROM", // id of present sender, no termination
	"TIS",  // id of present sender, signal termination; protocol continuation
	"CMD",  // ALE system-wide orderwire
	"REP"   // duplication of previous preamble with new data field
};

string ale_codec::command_strs[] = {
	"Advanced LQA",
	"LQA",
	"DBA",
	"Channels",
	"DTM",
	"Freq",
	"Mode selection",
	"Noise report"
};

char ale_codec::ASCII_Set[128] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

const int ale_codec::symbol_lookup[22] = {8,8,8,8,8,8,0,8,1,8,3,8,2,8,6,8,7,8,5,8,4,8};
const int ale_codec::vote_lookup[4]    = {0,0,1,1};

ale_codec::ale_codec() {
	mode = MODE_ALE;

	golay = new ale_golay;

	samplerate = ALE_SAMPLERATE;
	frequency = 1500;

// Create the twiddles etc for the FFT
	for (int i = 0; i < ALE_FFT_LEN; i++) {
		fft_cs_twiddle[i] = cos(-2.0 * M_PI * i / ALE_FFT_LEN);
		fft_ss_twiddle[i] = sin(-2.0 * M_PI * i / ALE_FFT_LEN);
		fft_history[i]    = 0;
	 }

	for (int i = 0; i < VOTE_ARRAY_LENGTH; i++)
		bits[i] = 0;

	for (int i = 0; i < VOTE_BUFFER_LENGTH; i++)
		majority_vote_array[i] = 0;

	for (int i = 0; i < ALE_FFT_LEN / 2; i++) {
		fft_mag[i] = 0;
		mag_sum[i] = 0;
		for (int j = 0; j < SYMBOLS_PER_WORD; j++)
			mag_history[i][j] = 0;
	}

	for (int i = 0; i < 8; i++) sig_filter[i].setLength(32);
	noise_filter.setLength(512);

	word_sync = false;
	noise_signal = 1.0;
	ale_signal = 0;

	input_buffer_pos = 0;
	fft_history_offset = 0;
	mag_history_offset = 0;

	last_symbol = 0;
	last_sync_position = 0;
	sample_count = 0;

	put_MODEstatus(MODE_ALE);
}

ale_codec::~ale_codec() {
	delete golay;
};

void ale_codec::init(void)
{
	word_sync = false;
	noise_signal = 1.0;
	ale_signal = 0;

	input_buffer_pos = 0;
	fft_history_offset = 0;
	mag_history_offset = 0;

	last_symbol = 0;
	last_sync_position = 0;
	sample_count = 0;

	ALE::continue_scanning();
}

//void ale_codec::decodeCMD (unsigned long word)
//{
//	int firstletter, metacmd;
//	firstletter = (word >> 14) & 0x0F;
//	metacmd     = (word >> 14) & 0x70;
//	if (metacmd == 6) {
//		printf (" %s ", command_strs[firstletter].c_str());
//	 }
//}

string ale_codec::decode_word (unsigned long word)
{
	unsigned char a, b, c;

	std::string msg;
	char szbuff[200];

	timestamp = time(NULL);
	UTC_time = gmtime(&timestamp);
	memset(szbuff, '\0', 200);
	snprintf (szbuff, sizeof(szbuff),
				"%02d:%02d:%02d,",
				UTC_time->tm_hour,
				UTC_time->tm_min,
				UTC_time->tm_sec);
	msg.assign(szbuff);

	c = word & 0x7F;
	b = (word >> 7) & 0x7F;
	a = (word >> 14) & 0x7F;
	preamble = (word >> 21) & 07;
	address = "   ";
	address[0] = a;
	address[1] = b;
	address[2] = c;

	if (preamble == CMD) {			// process CMD string
		command = a - '0';
		if (a >= 0 && a <= 7) {
			memset(szbuff, '\0', 200);
			snprintf(szbuff, sizeof(szbuff),
					"%s, %c%c%c%c,%s",
					preamble_strs[preamble].c_str(),
					preamble + '0', a, b, c,
					command_strs[command].c_str());
//					"%5s, '%c%c%c', %s, ",
//					preamble_strs[preamble].c_str(),
//					a, b, c,
//					command_strs[command].c_str());
			msg.append(szbuff);
		}
	} else							// check if ALE38 addr charset
		if ((ASCII_Set[a] == ASCII_38) && 
			(ASCII_Set[b] == ASCII_38) &&
			(ASCII_Set[c] == ASCII_38)) {

			command = -1;
			memset(szbuff, '\0', 200);
			snprintf (szbuff, sizeof(szbuff),
					"%s, %c%c%c%c",
					preamble_strs[preamble].c_str(),
					preamble + '0', a, b, c);
//						"%5s, '%c%c%c'",
//						preamble_strs[preamble].c_str(),
//						a, b, c);
			msg.append(szbuff);
	} else							// check if ALE64 addr charset
		if ((ASCII_Set[a] >= ASCII_64) && 
			(ASCII_Set[b] >= ASCII_64) &&
			(ASCII_Set[c] >= ASCII_64)) {
			command = -1;
			memset(szbuff, '\0', 200);
			snprintf (szbuff, sizeof(szbuff),
					"%6s, %c%c%c%c",
					preamble_strs[preamble].c_str(),
					preamble + '0', a, b, c);
//						"%5s, '%c%c%c'",
//						preamble_strs[preamble].c_str(),
//						a, b, c);
			msg.append(szbuff);
	} else {					// not ALE
		msg.clear();
	}
	return msg;
}

unsigned long ale_codec::modem_de_interleave_and_fec(int *input, int *errors)
{
	int i;

	worda = wordb = 0;

	for (i = 0; i < VOTE_BUFFER_LENGTH; ) {
		worda = input[i++] ? (worda << 1) + 1 : worda << 1 ;
		wordb = input[i++] ? (wordb << 1) + 1 : wordb << 1 ;
	}
	wordb = wordb ^ 0x000FFF;

// FEC, golay decode each word
	worda = golay->decode(worda, &error_a);
	wordb = golay->decode(wordb, &error_b);

// Return largest number of errors
	if (error_a > error_b)
		*errors = error_a;
	else
		*errors = error_b;

// Return single 24 bit word
	return (worda << 12) + wordb;
}
/*
  Process a new received symbol
*/
void ale_codec::modem_new_symbol(int sym)
{
	std::string rx_decode_str;

// Update the input bit buffer (149 bits)
	bits[input_buffer_pos] = (sym & 4) ? 1 : 0;

	if (++input_buffer_pos >= VOTE_ARRAY_LENGTH) input_buffer_pos -= VOTE_ARRAY_LENGTH;
	bits[input_buffer_pos] = (sym & 2) ? 1 : 0;

	if (++input_buffer_pos >= VOTE_ARRAY_LENGTH) input_buffer_pos -= VOTE_ARRAY_LENGTH;
	bits[input_buffer_pos] = (sym & 1) ? 1 : 0;

	if (++input_buffer_pos >= VOTE_ARRAY_LENGTH) input_buffer_pos -= VOTE_ARRAY_LENGTH;

// Do the majority vote, ignore stuff bits

	ber = 0;
	int p1, p2, p3;
	for (int i = 0; i < VOTE_BUFFER_LENGTH; i++) {
		p1 = i + input_buffer_pos;
		p2 = p1 + SYMBOLS_PER_WORD;
		p3 = p2 + SYMBOLS_PER_WORD;
		if (p1 >= VOTE_ARRAY_LENGTH) p1 -= VOTE_ARRAY_LENGTH;
		if (p2 >= VOTE_ARRAY_LENGTH) p2 -= VOTE_ARRAY_LENGTH;
		if (p3 >= VOTE_ARRAY_LENGTH) p3 -= VOTE_ARRAY_LENGTH;
		sum = bits[p1] + bits[p2] + bits[p3];
// all three repeats agree if sum is zero or three
		if (sum < 0) sum = 1;
		if (sum > 3) sum = 2;
		if (( sum == 1 ) || ( sum == 2))
			ber++;
		majority_vote_array[i] = vote_lookup[sum];
	}

// Test to see if it might be time for a word sync

	if (!word_sync) {
		if (ber <= BAD_VOTE_THRESHOLD) {
			word = modem_de_interleave_and_fec(majority_vote_array, &errors);
// If errors are less than or equal to allowable number of decode errors
// Decode the data word
			if (errors <= ale_config.decode_errors) {
				rx_decode_str = decode_word(word);
				word_sync = true;
				word_sync_position = input_buffer_pos;
//				dwell_on_channel();
			} else
				word_sync = false;
		}
	} else {
		if (input_buffer_pos == word_sync_position) {
// Signal new word
			word = modem_de_interleave_and_fec(majority_vote_array, &errors);
			if (errors <= ale_config.decode_errors) {
				rx_decode_str = decode_word (word);
				word_sync = true;
//			dwell_on_channel();
			} else
				word_sync = false;
		} else {
			word_sync = false;
		}
	}

// debugging & data output
//std::string sarray = "B";
//for (int i = 0; i < VOTE_ARRAY_LENGTH; i++)
//	sarray += '0' + bits[i];
//sarray += '\n';

//if (!rx_decode_str.empty()) {
//	std::string varray = "V";
//	for (int i = 0; i < SYMBOLS_PER_WORD - 1; i++)
//		varray += '0' + majority_vote_array[i];
//	size_t pos = sarray.find(varray.substr(1));
//	if (pos != std::string::npos)
//		varray.insert(1, std::string(pos - 1, '^'));
//	varray += '\n';
//	sarray.append(varray);

//	LOG_INFO("%s", msgout);
//	sarray.append(msgout).append("\n");
//	int n = 0;
//	while (msgout[n++]) { put_rx_char (msgout[n], FTextBase::RECV); }
//	put_rx_char( '\n', FTextBase::RECV);

if (word_sync) {
//	ALE::write_rxstr (std::string(msgout).append("\n"));
	ALE::write_rxstr(rx_decode_str, ber, sinad);
	ALE::parse_heard(preamble, address, ber, int(sinad));
}
//}

}

// Main Signal processing routine

int ale_codec::rx_process(const double *sample, int length)
{
	double mag;
	if (length == 0) return 0;
	for (int i = 0; i < length; i++) {

// Sliding FFT for signal bins only
		for (int n = 6; n <= 20; n += 2) {
			temp_real       = fft_out[n].real - fft_history[fft_history_offset] + sample[i];
			temp_imag       = fft_out[n].imag;
			fft_out[n].real = (temp_real * fft_cs_twiddle[n]) - (temp_imag * fft_ss_twiddle[n]);
			fft_out[n].imag = (temp_real * fft_ss_twiddle[n]) + (temp_imag * fft_cs_twiddle[n]);
			mag = sqrt((fft_out[n].real * fft_out[n].real) + (fft_out[n].imag * fft_out[n].imag) );
			fft_mag[n]      = sig_filter[(n - 6) / 2].run(mag);
		}
		fft_history[fft_history_offset] = sample[i];

// Save the min & max output values in a running average over one ALE word
// This is used for bit sync and symbol decoding
		max_magnitude = 0;
		max_offset    = 0;
		double noise = 0;

		for (int n = 6; n <= 20; n += 2) {
			if ( fft_mag[n] > max_magnitude ) {
				max_magnitude = fft_mag[n];
				max_offset    = n;
			}
		}
		for (int n = 6; n <=20; n += 2) {
			if (n != max_offset)
				noise = noise_filter.run(fft_mag[n]);
		}

		noise /= 7;

		if (noise <= 0) noise = 1e-6;
		if (max_magnitude <= 0) max_magnitude = 1e-8;
		double sn = 20 * log10(max_magnitude / noise);

		sinad = sn ;

// Update the magnitude history
		mag_sum[sample_count]                        -= mag_history[sample_count][mag_history_offset];
		mag_sum[sample_count]                        += max_magnitude;
		mag_history[sample_count][mag_history_offset] = max_magnitude;

// Update bit sync position
		if (!word_sync) {
			double max = 0;
			for (int n = 6; n <= 20; n += 2) {
				if (mag_sum[n] > max) {
					max = mag_sum[n];
					last_sync_position = n;
				}
			}
		}

// See what symbol has been received
		if (sample_count == last_sync_position) {
			last_symbol = symbol_lookup[max_offset];
		}

// Always output at the same position
		if (sample_count == 0) {
			modem_new_symbol(last_symbol);
		}

		fft_history_offset++;
		fft_history_offset = fft_history_offset % ALE_FFT_LEN;
		sample_count++;
		sample_count = sample_count % MOD_64;
		if (sample_count == 0) {
			mag_history_offset++;
			mag_history_offset = mag_history_offset % SYMBOLS_PER_WORD;
		}
	}
	return 0;
}

//----------------------------------------------------------------------
// tx processing
//
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// transmit octal word
// start and stop at 0 slope point in waveform
//----------------------------------------------------------------------

int ale_codec::oct_freq[] = {750, 1000, 1500, 1250, 2500, 2250, 1750, 2000};
double ale_codec::outbuf[64];

void ale_codec::xmt_ale_octet(int octet, int which)
{
	octet &= 0xFF;
	octet -= '0';
	if (octet < 0 || octet > 7) return;
	double k = oct_freq[octet] * 2.0 * M_PI / samplerate;

	memset ((void*)outbuf, 0, 64 * sizeof(double));
	for (int i = 0; i < 64; i++)
		outbuf[i] = cos (i * k);
	ModulateXmtr(outbuf, 64);
}

std::string ale_codec::binary_str(unsigned long nbr, int size)
{
	static std::string reverse;
	static std::string binary;
	binary.clear();
	reverse.clear();

	while (nbr) {
		reverse += (nbr & 1) ? '1' : '0';
		nbr /= 2;
	}
	while (reverse.length() < (size_t)size) reverse += '0';
	for (size_t n = reverse.length(); n > 0; n--) {
		binary += reverse[n-1];
	}
	return binary;
}

std::string ale_codec::octet_str(std::string binary)
{
	static std::string octets;
	octets.clear();
	for (size_t n = 0; n < binary.length(); n += 3) {
		octets += (binary[n]   - '0') * 4 +
				  (binary[n+1] - '0') * 2 +
				  (binary[n+2] - '0') + '0';
	}
	return octets;
}

//----------------------------------------------------------------------
// tx_fec
//
// convert 24 bit ale word to 147 bit golay-encoded, interleaved word
//
// input ale_word, 24 data bits
// 
// pppaaaaaaabbbbbbbccccccc
// ppp  - preamble
// a... - character 1
// b... - character 2
// c... - character 3
//

std::string ale_codec::tx_fec(unsigned long ale_word)
{
	unsigned long a = ale_word & 07777; // lower 12 bits
	unsigned long b = ale_word >> 12;  // upper 12 bits

// two 24-bit Golay code words
	unsigned long g0 = golay->encode(a);	// lower 12 bits
	unsigned long g1 = golay->encode(b);	// upper 12 bits
	g0 ^= 07777;							// invert lower check bits

// interleave the two golay encoded 12 bit words
// 24 bits becomes 48 bits

	unsigned long t0 = 1;
	unsigned long ta = 2;
	unsigned long tb = 1;
	unsigned long long txw = 0L;

	for (int i = 0; i < 24; i++) {
		if ((g1 & t0) == t0) txw |= ta;
		if ((g0 & t0) == t0) txw |= tb;
		t0 *= 2;
		ta *= 4;
		tb *= 4;
	}

// create 48 bit word string
	std::string ale_str = binary_str(txw, 48);

// triplicate the string and add stuff bits
	std::string Txw;
	Txw.assign(ale_str).append("0").append(ale_str).append("0").append(ale_str).append("0");

// convert to octal string format for use in modulator
	ale_str = octet_str(Txw);

	return ale_str;
}

void ale_codec::xmt_ale_triplet(std::string pre_string, std::string data_string)
{
	tx_preamble_string = pre_string;
	tx_string = data_string;

	unsigned long ale_word = 0L;

	tx_preamble = -1;
	for (int n = 0; n < 8; n++) {
		if (tx_preamble_string == preamble_strs[n]) {
			tx_preamble = n;
			break;
		}
	}
	if (tx_preamble == -1 ) return;

	if (tx_string.length() == 1) tx_string.append("@@");
	if (tx_string.length() == 2) tx_string.append("@");
	if (tx_string.length() > 3) tx_string.erase(3);
	tx_string[0] = toupper(tx_string[0]);
	tx_string[1] = toupper(tx_string[1]);
	tx_string[2] = toupper(tx_string[2]);
/*
std::cout << tx_preamble_string << ", " << tx_string[0] << ", " << tx_string[1] << ", " << tx_string[2] << std::endl;
std::cout << binary_str(tx_preamble, 3) << ", " 
		  << binary_str(tx_string[0], 7) << ", " 
		  << binary_str(tx_string[1], 7) << ", " 
		  << binary_str(tx_string[2], 7) << std::endl;
*/
	ale_word |= tx_preamble;
	ale_word <<= 7;
	ale_word |= tx_string[0];
	ale_word <<= 7;
	ale_word |= tx_string[1];
	ale_word <<= 7;
	ale_word |= tx_string[2];

	std::string octets = tx_fec(ale_word);

	size_t len = octets.length();

	xmt_ale_octet(octets[0], START);

	for (size_t n = 1; n < len - 1; n++)
		xmt_ale_octet(octets[n], MID);

	xmt_ale_octet(octets[len - 1], END);

	ALE::write_txstr(std::string(tx_preamble_string).append(" ").append(tx_string).append("\n"));

}

// this is just a dummy TX for testing purposes

int ale_codec::tx_process()
{
	int c = get_tx_char();

	if (c == GET_TX_CHAR_NODATA) {
		if (stopflag) {
			stopflag = false;
			put_echo_char('\n');
			tx_string.clear();
			return -1;
		}
		Fl::awake(); MilliSleep(10);
		return 0;
	}
	if (c == GET_TX_CHAR_ETX || stopflag) {
		return -1;
	}
/*
	tx_string += c;

	if (c != '\r') {
std::cout << (int)c << " : " << tx_string << std::endl;
		Fl::awake(); MilliSleep(10);
		return 0;
	}

// parse the transmit string into preamble and chars
std::cout << "tx_string: '" << tx_string << "'" << std::endl;

	size_t p = tx_string.find(' ');
	if (p == std::string::npos) {
		tx_string.clear();
		Fl::awake(); MilliSleep(10);
		return 0;
	}
	tx_preamble_string = tx_string.substr(0, p);
	for (size_t n = 0; n < tx_preamble_string.length(); n++)
		tx_preamble_string[n] = toupper(tx_preamble_string[n]);

std::cout << tx_preamble_string << std::endl;
*/

//MilliSleep(50);
//LOG_INFO("%s","Tx disabled");
//return -1;

double silence[2048];
memset ((void*)silence, 0, 2048 * sizeof(double));

	ModulateXmtr(silence, 2048);
	for (int i = 0; i < 10; i++) {
		xmt_ale_triplet("TWS", "W1H");
		xmt_ale_triplet("DATA", "KJ@");
		xmt_ale_triplet("REP", "KJ@");
	}
	ModulateXmtr(silence, 2048);
	for (int i = 0; i < 4; i++) {
		xmt_ale_triplet("TWS", "W1H");
	}
	for (int i = 0; i < 4; i++) {
		xmt_ale_triplet("DATA", "KJ@");
		xmt_ale_triplet("REP", "KJ@");
	}
	ModulateXmtr(silence, 2048);
	tx_string.clear();
//	return -1;

	return 0;
}

